@@ -24,7 +24,6 @@ import shutil
 import subprocess
 import sys
 import tarfile
-import tempfile
 import time
 
 from ironic_lib import disk_utils
@@ -35,6 +34,8 @@ from oslo_log import log as logging
 from oslo_serialization import base64
 from oslo_serialization import jsonutils
 from oslo_utils import units
+import requests
+import tenacity
 
 from ironic_python_agent import errors
 
@@ -77,6 +78,8 @@ DEVICE_EXTRACTOR = re.compile(r'^(?:(.*\d)p|(.*\D))(?:\d+)$')
 PARTED_TABLE_TYPE_REGEX = re.compile(r'^.*partition\s+table\s*:\s*(gpt|msdos)',
                                      re.IGNORECASE)
 
+_EARLY_LOG_BUFFER = []
+
 
 def execute(*cmd, **kwargs):
     """Convenience wrapper around ironic_lib's execute() method.
@@ -86,26 +89,6 @@ def execute(*cmd, **kwargs):
     return ironic_utils.execute(*cmd, **kwargs)
 
 
-def try_execute(*cmd, **kwargs):
-    """The same as execute but returns None on error.
-
-    Executes and logs results from a system command. See docs for
-    oslo_concurrency.processutils.execute for usage.
-
-    Instead of raising an exception on failure, this method simply
-    returns None in case of failure.
-
-    :param cmd: positional arguments to pass to processutils.execute()
-    :param kwargs: keyword arguments to pass to processutils.execute()
-    :raises: UnknownArgumentError on receiving unknown arguments
-    :returns: tuple of (stdout, stderr) or None in some error cases
-    """
-    try:
-        return execute(*cmd, **kwargs)
-    except (processutils.ProcessExecutionError, OSError) as e:
-        LOG.debug('Command failed: %s', e)
-
-
 def _read_params_from_file(filepath):
     """Extract key=value pairs from a file.
 
@@ -144,40 +127,33 @@ def _get_vmedia_device():
             pass
 
 
-@contextlib.contextmanager
-def _mounted(source):
-    """A context manager for a temporary mount."""
-    dest = tempfile.mkdtemp()
+def _find_vmedia_device_by_labels(labels):
+    """Find device matching any of the provided labels for virtual media"""
+    candidates = []
     try:
-        try:
-            execute("mount", source, dest)
-        except processutils.ProcessExecutionError as e:
-            msg = ("Unable to mount virtual media device %(device)s: "
-                   "%(error)s" % {'device': source, 'error': e})
-            raise errors.VirtualMediaBootError(msg)
+        lsblk_output, _e = execute('lsblk', '-p', '-P', '-oKNAME,LABEL')
+    except processutils.ProcessExecutionError as e:
+        _early_log('Was unable to execute the lsblk command. %s', e)
+        return
 
-        yield dest
-    finally:
-        try:
-            execute("umount", dest)
-        except processutils.ProcessExecutionError:
-            pass
+    for device in ironic_utils.parse_device_tags(lsblk_output):
+        for label in labels:
+            if label.upper() == device['LABEL'].upper():
+                candidates.append(device['KNAME'])
 
-        try:
-            shutil.rmtree(dest)
-        except Exception:
-            pass
-
-
-def _find_device_by_labels(labels):
-    """Find device matching any of the provided labels."""
-    for label in labels + [lbl.upper() for lbl in labels]:
-        try:
-            path, _e = execute('blkid', '-L', label)
-        except processutils.ProcessExecutionError:
-            pass
+    for candidate in candidates:
+        # We explicitly take the device and run it past _check_vmedia_device
+        # as there *can* be candidate entries, and we only want to return
+        # one that seems most likely to be the actual device, and the vmedia
+        # check code also evaluates the device overall, instead of just the
+        # block device with a label of some sort.
+        if _check_vmedia_device(candidate):
+            return candidate
         else:
-            return path.strip()
+            _early_log('Found possible vmedia candidate %s, however '
+                       'the device failed vmedia validity checking.',
+                       candidate)
+    _early_log('Did not identify any virtual media candidates devices.')
 
 
 def _get_vmedia_params():
@@ -187,19 +163,23 @@ def _get_vmedia_params():
     :raises: VirtualMediaBootError when it cannot find the virtual media device
     """
     parameters_file = "parameters.txt"
-    vmedia_device_file = _find_device_by_labels(['ir-vfd-dev'])
+    vmedia_device_file = _find_vmedia_device_by_labels(['ir-vfd-dev'])
     if not vmedia_device_file:
-        # TODO(rameshg87): This block of code is there only for compatibility
-        # reasons (so that newer agent can work with older Ironic). Remove
-        # this after Liberty release.
+        # This falls back to trying to find a matching device by name/type.
+        # if not found, it is likely okay to just fail out and treat it as
+        # No device found as there are multiple ways to launch IPA, and all
+        # vmedia styles should be treated consistently.
         vmedia_device = _get_vmedia_device()
         if not vmedia_device:
-            msg = "Unable to find virtual media device"
-            raise errors.VirtualMediaBootError(msg)
+            return {}
 
         vmedia_device_file = os.path.join("/dev", vmedia_device)
 
-    with _mounted(vmedia_device_file) as vmedia_mount_point:
+        if not _check_vmedia_device(vmedia_device_file):
+            # If the device is not valid, return an empty dictionary.
+            return {}
+
+    with ironic_utils.mounted(vmedia_device_file) as vmedia_mount_point:
         parameters_file_path = os.path.join(vmedia_mount_point,
                                             parameters_file)
         params = _read_params_from_file(parameters_file_path)
@@ -209,7 +189,16 @@ def _get_vmedia_params():
 
 def _early_log(msg, *args):
     """Log via printing (before oslo.log is configured)."""
-    print('ironic-python-agent:', msg % args, file=sys.stderr)
+    log_entry = msg % args
+    _EARLY_LOG_BUFFER.append(log_entry)
+    print('ironic-python-agent:', log_entry, file=sys.stderr)
+
+
+def log_early_log_to_logger():
+    """Logs early logging events to the configured logger."""
+
+    for entry in _EARLY_LOG_BUFFER:
+        LOG.info("Early logging: %s", entry)
 
 
 def _copy_config_from(path):
@@ -245,22 +234,107 @@ def _find_mount_point(device):
         return path.strip()
 
 
+def _check_vmedia_device(vmedia_device_file):
+    """Check if a virtual media device appears valid.
+
+    Explicitly ignores partitions, actual disks, and other itmes that
+    seem unlikely to be virtual media based items being provided
+    into the running operating system via a BMC.
+
+    :param vmedia_device_file: Path to the device to examine.
+    :returns: False by default, True if the device appears to be
+              valid.
+    """
+    try:
+        output, _e = execute('lsblk', '-n', '-s', '-P', '-b',
+                             '-oKNAME,TRAN,TYPE,SIZE',
+                             vmedia_device_file)
+    except processutils.ProcessExecutionError as e:
+        _early_log('Failed to execute lsblk. lsblk is required for '
+                   'virtual media identification. %s', e)
+        return False
+    try:
+        for device in ironic_utils.parse_device_tags(output):
+            if device['TYPE'] == 'part':
+                _early_log('Excluding device %s from virtual media'
+                           'consideration as it is a partition.',
+                           device['KNAME'])
+                return False
+            if device['TYPE'] == 'rom':
+                # Media is a something like /dev/sr0, a Read only media type.
+                # The kernel decides this by consulting the underlying type
+                # registered for the scsi transport and thus type used.
+                # This will most likely be a qemu driven testing VM,
+                # or an older machine where SCSI transport is directly
+                # used to convey in a virtual
+                return True
+            if device['TYPE'] == 'disk' and device['TRAN'] == 'usb':
+                # We know from experience on HPE machines, with ilo4/5, we see
+                # and redfish with edgeline gear, return attachment from
+                # pci device 0c-03.
+                # https://linux-hardware.org/?probe=4d2526e9f4
+                # https://linux-hardware.org/?id=pci:103c-22f6-1590-00e4
+                #
+                # Dell hardware takes a similar approach, using an Aten usb hub
+                # which provides the standing connection for the BMC attached
+                # virtual kvm.
+                # https://linux-hardware.org/?id=usb:0557-8021
+                #
+                # Supermicro also uses Aten on X11, X10, X8
+                # https://linux-hardware.org/?probe=4d0ed95e02
+                #
+                # Lenovo appears in some hardware to use an Emulux Pilot4
+                # integrated hub to proivide device access on some hardware.
+                # https://linux-hardware.org/index.php?id=usb:2a4b-0400
+                #
+                # ??? but the virtual devices appear to be American Megatrends
+                # https://linux-hardware.org/?probe=076bcef32e
+                #
+                # Fujitsu hardware is more uncertian, but appears to be similar
+                # in use of a USB pass-through
+                # http://linux-hardware.org/index.php?probe=cca9eab7fe&log=dmesg
+                if device['SIZE'] != "" and int(device['SIZE']) < 4294967296:
+                    # Device is a usb backed block device which is smaller
+                    # than 4 GiB
+                    return True
+                else:
+                    _early_log('Device %s appears to not qualify as virtual '
+                               'due to the device size. Size: %s',
+                               device['KNAME'], device['SIZE'])
+            _early_log('Device %s was disqualified as virtual media. '
+                       'Type: %s, Transport: %s',
+                       device['KNAME'], device['TYPE'], device['TRAN'])
+        return False
+    except KeyError:
+        return False
+
+
+def _booted_from_vmedia():
+    """Indicates if the machine was booted via vmedia."""
+    params = _read_params_from_file('/proc/cmdline')
+    return params.get('boot_method') == 'vmedia'
+
+
 def copy_config_from_vmedia():
     """Copies any configuration from a virtual media device.
 
     Copies files under /etc/ironic-python-agent and /etc/ironic-python-agent.d.
     """
-    vmedia_device_file = _find_device_by_labels(
+    vmedia_device_file = _find_vmedia_device_by_labels(
         ['config-2', 'vmedia_boot_iso'])
     if not vmedia_device_file:
         _early_log('No virtual media device detected')
         return
-
+    if not _booted_from_vmedia():
+        _early_log('Cannot use configuration from virtual media as the '
+                   'agent was not booted from virtual media.')
+        return
+    # Determine the device
     mounted = _find_mount_point(vmedia_device_file)
     if mounted:
         _copy_config_from(mounted)
     else:
-        with _mounted(vmedia_device_file) as vmedia_mount_point:
+        with ironic_utils.mounted(vmedia_device_file) as vmedia_mount_point:
             _copy_config_from(vmedia_mount_point)
 
 
@@ -404,7 +478,6 @@ def get_command_output(command):
     except (processutils.ProcessExecutionError, OSError) as e:
         error_msg = ('Failed to get the output of the command "%(command)s". '
                      'Error: %(error)s' % {'command': command, 'error': e})
-        LOG.error(error_msg)
         raise errors.CommandExecutionError(error_msg)
     return io.BytesIO(out)
 
@@ -473,7 +546,7 @@ def collect_system_logs(journald_max_lines=None):
         try:
             io_dict[file_name] = get_command_output(command)
         except errors.CommandExecutionError:
-            pass
+            LOG.debug('Collecting logs from command %s has failed', command)
 
     io_dict = {}
     file_list = []
@@ -811,3 +884,42 @@ def create_partition_table(dev_name, partition_table_type):
         msg = "Failed to create partition table on {}: {}".format(
             dev_name, e)
         raise errors.CommandExecutionError(msg)
+
+
+class StreamingClient:
+    """A wrapper around HTTP client with TLS, streaming and error handling."""
+
+    _CHUNK_SIZE = 1 * units.Mi
+
+    def __init__(self, verify_ca=True):
+        if verify_ca:
+            self.verify, self.cert = get_ssl_client_options(CONF)
+        else:
+            self.verify, self.cert = False, None
+
+    @contextlib.contextmanager
+    def __call__(self, url):
+        """Execute a GET request and start streaming.
+
+        :param url: Target URL.
+        :return: A generator yielding chunks of data.
+        """
+        @tenacity.retry(
+            retry=tenacity.retry_if_exception_type(requests.ConnectionError),
+            stop=tenacity.stop_after_attempt(
+                CONF.image_download_connection_retries + 1),
+            wait=tenacity.wait_fixed(
+                CONF.image_download_connection_retry_interval),
+            reraise=True)
+        def _get_with_retries():
+            return requests.get(url, verify=self.verify, cert=self.cert,
+                                stream=True,
+                                timeout=CONF.image_download_connection_timeout)
+
+        try:
+            with _get_with_retries() as resp:
+                resp.raise_for_status()
+                yield resp.iter_content(self._CHUNK_SIZE)
+        except requests.RequestException as exc:
+            raise errors.CommandExecutionError(
+                "Unable to read data from %s: %s" % (url, exc))
